#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

package Expr;
use v5.014;
use strict;
use warnings;
use Carp;

use lib '.';
use parent 'Object';
use POSIX;
use Math::Trig;


my $float_patt = '\d+\.?(\d*)([eE][+-]?\d+)?';

my %predefined = (abs          => 'abs',
				  root         => 'sqrt',
				  floor        => 'floor',
				  ceil         => 'ceil',
				  log          => 'log',
				  log10        => 'log10',
				  exp          => 'exp',
				  sin          => 'sin',
				  cos          => 'cos',
				  tan          => 'tan',
				  sec          => 'sec',
				  csc          => 'csc',
				  cot          => 'cotan',
				  sinh         => 'sinh',
				  cosh         => 'cosh',
				  tanh         => 'tanh',
				  sech         => 'sech',
				  csch         => 'csch',
				  coth         => 'cotanh',
				  asin         => 'asin',
				  acos         => 'acos',
				  atan         => 'atan',
				  arcsec       => 'asec',
				  arccsc       => 'acsc',
				  arccot       => 'acotan',
				  arcsinh      => 'asinh',
				  arccosh      => 'acosh',
				  arctanh      => 'atanh',
				  arcsech      => 'asech',
				  arccsch      => 'acsch',
				  arccoth      => 'acotanh');

sub new {
	my ($class, $state, $desc) = @_;

	my $self = $class->SUPER::new(state => $state,
								  desc  => $desc);
	bless $self, $class;
}


sub value {
	my $self = shift;
	my $desc = $self->desc;
	$desc =~ s/\s+//g;

	my @s;
	while ($desc) {
		$desc =~ /^(([-+*\/%^()])|(\[[^\]]+\](_0)?)|([a-z_][a-z0-9_]*)|($float_patt))(.*?)$/i;
		push @s, $1;
		$desc = $9;
	}

	for (my $i = 0; $i < scalar @s; $i++) {
		if ($s[$i] =~ /^[a-z_][a-z0-9_]*$/i
			&& ($i == scalar @s - 1 || $s[$i + 1] ne '(')) {
			my $c = $self->state->consts->find($s[$i]);
			croak "undefined constant '$s[$i]'" unless defined $c;
			$s[$i] = ref($c->{value}) ? $c->{value}->value : $c->value;
		} elsif ($s[$i] =~ /^\[[^\]]+\](_0)?$/) {
			$s[$i] =~ s/_0$//;
			my $c = $self->state->species->find($s[$i]);
			croak "undefined species '$s[$i]'" unless defined $c;
			$s[$i] = $c->concentration;
		} elsif (   $s[$i] !~ /^(([-+*\/%\^()])|($float_patt))$/
				 && $i < scalar @s - 1 && $s[$i + 1] eq '(') {
			for (keys %predefined) {
				if ($s[$i] =~ /^$_$/i) {
					$s[$i] = $predefined{$_};
					last;
				}
			}
		} elsif ($s[$i] eq '^') {
			$s[$i] = '**';
		}
	}

	my $expr = join '', @s;
	eval $expr or croak   "expression '$self->desc' (cpnverted to '$expr') "
		                . "does not compute";
}


1;
