#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

package Function;
use v5.014;
use strict;
use warnings;
use Carp;

use lib '.';
use parent 'Object';
use Func_Param;

my $float_patt = '\d+\.?(\d*)([eE][+-]?\d+)?';

# Names of all nmathematical functions and constants known by COPASI
# that can be used in functions (if/IF and logical operators are missing!) 

my @predefined = qw/abs
                    root
                    floor
                    ceil
                    factorial
                    log
                    log10
                    exp
                    sin
                    cos
                    tan
                    sec
                    csc
                    cot
                    sinh
                    cosh
                    tanh
                    sech
                    csch
                    coth
                    asin
                    acos
                    atan
                    arcsec
                    arccsc
                    arccot
                    arcsinh
                    arccosh
                    arctanh
                    arcsech
                    arccsch
                    arccoth
                    uniform
                    normal
                    pi
                    exponentiale
                    true
                    false
                    infinity/;


# Creates a new function

sub new {
	my ($class, $state, $str) = @_;

	# The line describing a function may start with a name for the function,
	# followed by a colon. See if we got one (otherwise functions are named
	# 'R0', 'R1' etc.)

	my ($name, $rem) = split /\s*:\s*/, $str;
	my ($formula, $rev, $type) = split /\s*,\s*/, $rem;

	my $is_rev;
	if (! defined $rev) {
		$is_rev = 'unspecified';
	} elsif ($rev =~ /^rev(ersible)?$/) {
		$is_rev = 'true';
	} elsif ($rev =~ /^irrev(ersible)?$/) {
		$is_rev = 'false';
	} else {
		croak "invalid keyword '$rev'\n";
	}

	# Remove all spaces except from species names in initial concentrations

	$formula = join '', map {s/\s+//g unless /^\[/; $_}
						split /\s*(\[[^\]]+\]_0)\s*/, $formula;
	
	my $self = $class->SUPER::new(state      => $state,
								  name       => $name,
								  formula    => $formula,
								  reversible => $is_rev,
								  type       => defined $type ?
								                $type : "UserDefined");

	$self->state->functions->add($self);
	Func_Param->new($state, $self, $_) for _analyze($formula);

	return $self;
}


#

sub check {
	my $self = shift;

	my $formula = $self->formula;
	while ($formula =~ /\[([^\]]+)\]_0/) {
		die   "unknown species '$1' found in formula for function '"
			. $self->name . "'\n"
			unless defined $self->state->species->get($1);
		$formula =~ s/\[([^\]]+)\]_0//;
	}
}


# Returns if the function was marked as reversible

sub is_reversible {
	shift->{reversible} ne 'false';
}


# Return reference to array of all parameters of the function

sub parameter {
	my $self = shift;
	$self->state->func_params->all_of_function($self);
}


# Return reference to array of all substrates of the function

sub substrates {
	my $self = shift;
	[grep {$_->type eq 'substrate'} @{$self->parameter}]
}


# Return reference to array of all products of the function

sub products {
	my $self = shift;
	[grep {$_->type eq 'product'} @{$self->parameter}]
}


# Return reference to array of all constants used in the function

sub constants {
	my $self = shift;
	[grep {$self && $_->type eq 'constant'} @{$self->parameter}]
}


# Return string with COPASI XML code for the function

sub emit {
	my ($self, $indent) = @_;
	my $s = sprintf   "%s<Function key=\"Function_%d\" "
		            . "name=\"%s\" "
					. "type=\"%s\" "
					. "reversible=\"%s\">\n",
                    $indent, $self->state->functions->index($self),
					$self->name, $self->type, $self->reversible;
	$indent .= '  ';
	$s .= "$indent<Expression>\n"; 
	$indent .= '  ';

	# Replace initial concentrations by the corresponding numeric values
	# before writing out the formula

	my $formula = $self->formula;
	$formula =~
			s/\[([^\]]+)\]_0/$self->state->species->get($1)->concentration/eg;
	$s .= $indent . $formula . "\n";;

	$indent =~ s/  //;
	$s .= "$indent</Expression>\n";
	$s .= "$indent<ListOfParameterDescriptions>\n";
	$indent .= '  ';
	$s .= $_->emit($indent) for @{$self->parameter};
	$indent =~ s/  //;
	$s .= "$indent</ListOfParameterDescriptions>\n";
	$indent =~ s/  //;
	$s .= "$indent</Function>\n";
}


sub emit_reaction {
	my ($self, $reaction, $indent) = @_;
	my $s = sprintf   "%s<KineticLaw function=\"Function_%d\" "
		            . "unitType=\"Default\" "
					. "scalingCompartment=\"CN=Root,Model=%s,"
					. "Vector=Compartments[compartment]\">\n",
					$indent, $self->state->functions->index($self),
					$self->state->model->name;

	$indent .= '  ';
	$s .= "$indent<ListOfCallParameters>\n";

	$indent .= '  ';

	my @call_params = @{$self->parameter};
	my $reaction_constants = $reaction->constants;
	my @reaction_substrates;
	for my $s (@{$reaction->substrates}) {
		for my $i (1 ... $s->{stoichiometry}) {
			push @reaction_substrates, $s;
		}
	}

	my @reaction_products;
	for my $p (@{$reaction->products}) {
		for my $i (1 ... $p->{stoichiometry}) {
			push @reaction_products, $p;
		}
	}

	for my $p (@call_params) {
		$s .= sprintf   "$indent<CallParameter "
			          . "functionParameter=\"FunctionParameter_%d\">\n",
			$self->state->func_params->index($p);
		$indent .= '  ';

		if ($p->type eq 'constant') {
			my $rc;
			for (@$reaction_constants) {
				if ($_->name eq $p->name) {
					$rc = $_;
					last;
				}
			}

			$s .= sprintf   "$indent<SourceParameter "
				          . "reference=\"Parameter_%d\"/>\n",
				          $self->state->params->index($rc);
		} elsif ($p->type eq 'substrate') {
			my $fi = $self->state->functions->index($self);
			if ($fi < 2) {
				for (@reaction_substrates) {
					$s .= sprintf   "$indent<SourceParameter "
						          . "reference=\"Metabolite_%d\"/>\n",
						          $self->state->species->index($_->{species});
				}
			} elsif (@reaction_substrates) {
				$s .= sprintf   "$indent<SourceParameter "
						      . "reference=\"Metabolite_%d\"/>\n",
					          $self->state->species->index(
								            $reaction_substrates[0]->{species});
				shift @reaction_substrates;
			}
		} elsif ($p->type eq 'product') {
			my $fi = $self->state->functions->index($self);
			if ($fi == 1) {
				for (@reaction_products) {
					$s .= sprintf   "$indent<SourceParameter "
						          . "reference=\"Metabolite_%d\"/>\n",
						          $self->state->species->index($_->{species});
				}
			} elsif (@reaction_products) {
				$s .= sprintf   "$indent<SourceParameter "
						      . "reference=\"Metabolite_%d\"/>\n",
					          $self->state->species->index(
								            $reaction_products[0]->{species});
				shift @reaction_products;
			}
		}

		$indent =~ s/  //;
		$s .= "$indent</CallParameter>\n";
	}

	$indent =~ s/  //;
	$s .= "$indent</ListOfCallParameters>\n";

	$indent =~ s/  //;
	$s .= "$indent</KineticLaw>\n";
}       


sub _analyze {
	# Remove initial concentrations, split at operators, then weed out
	# numbers, predefined constants, COPASI functions and repeats of items

	my $func = shift;
	$func =~ s/PRODUCT&lt;((substrate)|(product))_[ij]>/$1/g;
	$func =~ s/\[[^\]]+\]_0//a;
	my @list = split /\s*[\+\-\*\/\%\^\(\)]+\s*/, $func;
	@list = do {my %seen; grep {   ! /^$float_patt$/
								&& ! _is_predefined($_)
								&& ! $seen{$_}++} @list};
}


# Returns if an element of a function is something predefined by COPASI

sub _is_predefined {
	my $key = shift;
	!! grep {$key =~ /^$_$/i} @predefined;
}

1;
