#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

package Plot;
use v5.014;
use strict;
use warnings;
use Carp;

use lib '.';
use parent 'Object';


my $float_patt = '\d+\.?(\d*)([eE][+-]?\d+)?';


sub new {
	my ($class, $state, $type, $title, $log_x_scale, $log_y_scale) = @_;
	my $self = $class->SUPER::new(state       => $state,
								  type        => $type,
								  title       => $title,
								  log_x_scale => $log_x_scale,
								  log_y_scale => $log_y_scale,
								  species => []);
	$state->plots->add($self);
	return $self;
}


sub add_species {
	my ($self, $line) = @_;
	my $species = undef;

	for (@{$self->state->species->list}) {
		if ($line eq $_->name) {
			push @{$self->species}, $_;
			return;
		}
	}

	die "Unknown species '$line' to plot";
}


sub emit {
	my ($self, $indent) = @_;

	$self->{title} = sprintf "Plot%d", $self->state->plots->index($self)
		unless defined $self->title && $self->title ne '';

	push @{$self->species}, grep({$_->type ne 'fixed'
								  && (   $self->type ne 'Rate'
									  || ref($_) ne 'Species_Assignment')}
								 @{$self->state->report->species})
		if @{$self->species} == 0;

	my $s =   "$indent<PlotSpecification name=\"" . $self->title . "\" "
		    . "type=\"Plot2D\" active=\"1\" taskTypes=\"\">\n";
	$indent .= '  ';

	$s .= "$indent<Parameter name=\"log X\" type=\"bool\" value=\"0\"/>\n";
	$s .= "$indent<Parameter name=\"log Y\" type=\"bool\" value=\"0\"/>\n";
	$s .= "$indent<ListOfPlotItems>\n";
	$indent .= '  ';

	for (@{$self->species}) {
		$s .= sprintf "%s<PlotItem name=\"[%s]|Time\" type=\"Curve2D\">\n",
		              $indent, $_->name;
		$indent .= '  ';

		$s .=   "$indent<Parameter name=\"Line type\" "
			  . "type=\"unsignedInteger\" value=\"0\"/>\n";
		$s .=   "$indent<Parameter name=\"Line subtype\" "
			  . "type=\"unsignedInteger\" value=\"0\"/>\n";
		$s .=   "$indent<Parameter name=\"Line width\" "
			  . "type=\"unsignedFloat\" value=\"1\"/>\n";
		$s .=   "$indent<Parameter name=\"Symbol subtype\" "
			  . "type=\"unsignedInteger\" value=\"0\"/>\n";
		$s .=   "$indent<Parameter name=\"Color\" "
			  . "type=\"string\" value=\"auto\"/>\n";
		$s .=   "$indent<Parameter name=\"Recording Activity\" "
			  . "type=\"string\" value=\"during\"/>\n";
		$s .= "$indent<ListOfChannels>\n";
		$indent .= '  ';

		$s .= sprintf   "%s<ChannelSpec cn=\"CN=Root,Model=%s,"
			          . "Reference=Time\"/>\n",
			          $indent, $self->state->model->name;
		$s .= sprintf   "%s<ChannelSpec cn=\"CN=Root,Model=%s,"
			          . "Vector=Compartments[compartment],"
			          . "Vector=Metabolites[%s],Reference=%s\"/>\n",
					  $indent, $self->state->model->name, $_->name, $self->type;
		$indent =~ s/  //;
		$s .= "$indent</ListOfChannels>\n";
		$indent =~ s/  //;
		$s .= "$indent</PlotItem>\n";
	}

	$indent =~ s/  //;
	$s .= "$indent</ListOfPlotItems>\n";

	$indent =~ s/  //;
	$s .= "$indent</PlotSpecification>\n";
}


sub get_title {
	my ($self, $filename) = @_;

	my $title = $filename;
	$title =~ s/^.*?([^\/]\\]+)$/$1/;
	if ($title =~ /_((\[.+?\])|(.*?\(.*?\)))=$float_patt/) {
		$title =~ s/^.*?(_((\[.*?\])|(.*?\(.*?\)))=.*)$/$1/;
		$title =~ s/_((\[.*?\])|(.*?\(.*?\)))=/, $1=/;
		$title =~ s/=($float_patt)/ = $1/g;
		$title =~ s/(= $float_patt),(\S)/$1, $4/g;
		$title = $self->title . $title;
	} else {
		$title = $self->title;
	}
}


sub gnuplot_plot {
	my ($self, $filename, $gnuplot) = @_;
	my $state = $self->state;

	my $cmd =   "set term jpeg enhanced size 1024,800\n"
		      . "set output '${filename}_" . $self->title . ".jpg'\n"
			  . "set title '" . $self->get_title($filename) . "'\n"
			  . "set key right center\n"
			  . "set xlabel 'Time / s'\n"
			  . 'set ylabel '
			  . ($self->type eq 'Concentration' ?
				 "'[species] / mol L^{-1}'" : "'Rate / mol L^{-1} s^{-1}'")
			  . "\n"
			  . "aGet(name, i) = value(sprintf(\"_%s_%i\", name, i))\n"
			  . "aSet(name, i, val) = sprintf(\"_%s_%i = '%s'\", name, "
			  . "i, val)\n"
			  . "toNum(i) = i + 0\n";
	$cmd .= "set logscale x\nset format x '%g'\n" if $self->log_x_scale;
	$cmd .= "set logscale y\nset format y '%g'\n" if $self->log_y_scale;

	my @report_species = $self->state->report->species;
	my @species;
	my $list = '';
	my $offset =
		2 + ($self->type eq 'Concentration' ? 0 : @{$state->report->species});
	for (@{$self->species}) {
		my $index = $state->report->index($_, $self->type eq 'Rate');
		if ($index < 0) {
			say   "WARNING: species '" . $_->name . "' is not in report, will "
				. "also be missing in plot!";
		} else {
			push @species, $_;
			$list .= ($offset + $index) . " ";
		}
	}
			
	$list =~ s/\s+$//;
	$cmd .= "list = '$list'\n";

	my @indices = split /\s+/, $list;
	for (@species) {
		my $name = $_->name;
		$name =~ s/\_/\\_/g;
		$name =~ s/\^/\\^/g;
		$cmd .= "eval aSet('label', " . (shift @indices) . ", '[$name]')\n";
	}

	$cmd .=   "plot for [i in list] '$filename.crp' every ::1 "
		    . "using 1:toNum(i) w l lw 2 title aGet('label', toNum(i))\n";

	open my $fh, "|-", "$gnuplot"
		or die "Can't start '$gnuplot'\n";
	print $fh $cmd;
	close $fh;
}


1;
