#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

package Range;
use v5.014;
use strict;
use warnings;
use Carp;

use lib '.';
use parent 'Object';

my $float_patt = '\d+\.?(\d*)([eE][+-]?\d+)?';


sub new {
	my ($class, $parent, $val) = @_;
	my $self = $class->SUPER::new(parent   => $parent,
								  iterator => $parent->state->iterator,
								  list     => _make_list($val));
	$self->{iterator}->add($self);
	return $self;
}


# Return string to be used in file names for the range

sub desc {
	my $self = shift;
	return sprintf "%s=%.6g", $self->parent->desc, $self->value;
}


# Return number of elements of the range

sub count {
	return scalar @{shift->list};
}


# Return that value of range currently indexed by the iterator

sub value {
	my $self = shift;
	return $self->list->[$self->iterator->current_index($self)];
}


# Extract range values from inpuy line

sub _make_list {
	my $vals = shift;
	my @val_list;
	
	if ($vals =~ /:/) {         # list with description of values
		my $type = 'lin';
		$vals =~ s/^\s+//;
		$vals =~ s/\s+$//;
		my @v = split /\s*:\s*/, $vals;
		croak "invalid number of elements in list\n" if @v < 3 || @v > 4;

		if (@v == 3) {
			my @le = split /\s+/, $v[2];
			croak "unexpected content at end of value list\n"
				if    @le > 2
				   || (defined $le[1] && $le[1] ne 'lin' && $le[1] ne 'log');
			$v[2] = $le[0];
			push @v, defined $le[1] ? $le[1] : 'lin';
		}

		if (@v == 4) {
			if ($v[3] =~ /^(lin)|(log)$/) {
				$type = pop @v;
			} else {
				croak "unexpected identifier '$v[3]' at end of list\n";
			}
		}

		croak "start is not a number\n" unless $v[0] =~ /^$float_patt$/;
		croak "end is not a number\n" unless $v[1] =~ /^$float_patt$/;
		croak "start and end value are identical\n" if $v[0] == $v[1];
		croak "count is not an integer\n" unless $v[2] =~ /^\d+/;
		croak "count must be at least 2\n" unless  $v[2] > 1;

		if ($type eq 'lin') {
			for my $i (0..$v[2] - 1) {
				push @val_list, $v[0] + $i * ($v[1] - $v[0]) / ($v[2] - 1);
			}

		} else {
			my $start = log($v[0]) / log(10);
			my $end   = log($v[1]) / log(10);
			for my $i (0..$v[2] - 1) {
				push @val_list,
					 10**($start + $i * ($end - $start) / ($v[2] - 1));
			}
		}
	} else {                       # simple list of values
		@val_list = split /\s+/, $vals;
		croak "less than two elements in value list\n" if @val_list < 2;
		for (@val_list) {
			croak "invalid element '$_' in value list\n"
				unless $_ =~ /^($float_patt)$/;
		}
	}

	return \@val_list;
}
