#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

package Reaction;
use v5.014;
use strict;
use warnings;
use Carp;

use lib '.';
use parent 'Object';
use Param;


sub new {
	my ($class, $state) = @_;
	my $self = $class->SUPER::new(state      => $state,
								  name       =>'R' . $state->reactions->count,
								  substrates => [],
								  products   => [],
								  constants  => [],
								  function   => undef);

	$state->reactions->add($self);
	$self->_analyze($_[2]);

	return $self;
}


# Return string with COPASI XML code for reaction

sub emit {
	my ($self, $indent) = @_;
	my $r_index = $self->state->reactions->index($self);
	my $s = sprintf   "%s<Reaction key=\"Reaction_%d\" "
		            . "name=\"%s\" "
					. "reversible=\"%s\" "
					. "fast=\"false\" "
					. "addNoise=\"false\">\n",
					$indent, $self->state->reactions->index($self), $self->name,
		            $self->function->is_reversible ? 'true' : 'false';
	$indent .= '  ';
	$s .= "$indent<ListOfSubstrates>\n";
	$indent .= '  ';
	$s .= sprintf   "%s<Substrate metabolite=\"Metabolite_%d\" "
		          . "stoichiometry=\"%d\"/>\n",
		          $indent, $self->state->species->index($_->{species}),
		          $_->{stoichiometry}
		      for @{$self->substrates};
	$indent =~ s/  //;
	$s .= "$indent</ListOfSubstrates>\n";

	$s .= "$indent<ListOfProducts>\n";
	$indent .= '  ';
	$s .= sprintf   "%s<Product metabolite=\"Metabolite_%d\" "
		          . "stoichiometry=\"%d\"/>\n",
		          $indent, $self->state->species->index($_->{species}),
		          $_->{stoichiometry}
		      for @{$self->products};
	$indent =~ s/  //;
	$s .= "$indent</ListOfProducts>\n";

	$s .= "$indent<ListOfConstants>\n";
	$indent .= '  ';

	$s .= $_->emit($indent) for @{$self->state->params->all_of_reaction($self)};

	$indent =~ s/  //;
	$s .= "$indent</ListOfConstants>\n";

	$s .= $self->function->emit_reaction($self, $indent);

	$indent =~ s/  //;
	$s .= "$indent</Reaction>\n";
}


# Parses the complete line for the reaction

sub _analyze {
	my ($self, $desc) = @_;

	# If there's a ':' before the first ',' take what's before it as the
	# name of the reaction. Otherwise name the reaction as "R" followed
	# by the number of reactions already found.

	if ($desc =~ /^([^,:]+):\s*(.*)$/) {
		$self->{name} = $1;
		$desc = $2;
	}

	# Split the rest of the line at commas. The first part is supposed
	# to be the part that describes the reaction, the rest the (optional)
	# name of the function for the reaction, followed by constant names
	# and assigned values (or value ranges).

	my @args = split /\s*,\s*/, $desc;

	my $cnts = $self->_analyze_reaction(shift @args);
	$self->_analyze_params($self->_analyze_function(@args));

	# Make sure the number of substrates on the left hand side of the
	# reaction description isn't smaller than the substrates used in
	# the function for the reaction - and the same condition must be
	# satisfied for the number of products on the right hand side.

	croak "not enough substrates for function '${\$self->function->name}'\n"
		if $cnts->[0] < @{$self->function->substrates};
	croak "not enough products for function '${\$self->function->name}'\n"
		if $cnts->[1] < @{$self->function->products};
}


# Parses the part of the line for the reaction itself

sub _analyze_reaction {
	my ($self, $desc) = @_;

	# Split the reaction part at the '->'

	my ($substrate_side, $product_side) = split /\s*->\s*/, $desc;

	# There must be some text on both sides

	croak "invalid reaction\n"
		if $substrate_side eq '' || $product_side eq '';

	# Parse both sides

	return [$self->_analyze_side($self->substrates, $substrate_side),
			$self->_analyze_side($self->products, $product_side)];
}


# Parse one side of a reaction

sub _analyze_side {
	my ($self, $list, $side) = @_;
	my $count;

	# Remove leading and trailing white space and split at '+' (surrounded
	# by white space).

	$side =~ s/^\s*(.$?)\s*$/$1/;
	my @reactands = split /\s+\+\s+/, $side;

	# Deal with each of these snippets

	for (@reactands) {
		# Split at '*' (surrounded by spaces - this must result in no more
		# than two parts (thus "A" or "2 * A" or "A * 2" are ok, where 'A'
		# is a species name).

		my @comp = split /\s+\*\s+/, $_;
		croak "invalid reaction\n" if @comp > 2;

		# Handle the three possible cases

		my $species = $self->state->species->get($comp[0]);
	
		# Case of 'A' only

		if (@comp == 1) {
			croak "invalid species '$comp[0]' in reaction\n"
				unless defined $species;
			push @$list, {species       => $species,
						  stoichiometry => 1};
			$count++;
			$species->is_used(1);
			next;
		}

		# Case of 'A * 2'

		if (defined $species && $comp[1] =~ /^\d+$/) {
			push @$list, {species       => $species,
						  stoichiometry => $comp[1]};
			$count += $comp[1];
			$species->is_used(1);
			next;
		}

		# Case of '2 * A'

		$species = $self->state->species->get($comp[1]);
		if (defined $species && $comp[0] =~ /^\d+$/) {
			push @$list, {species       => $species,
						  stoichiometry => $comp[0]};
			$count += $comp[0];
			$species->is_used(1);
			next;
		}

		croak "invalid reacrion\n";
	}

	return $count;
}


sub _analyze_function {
	my ($self, @args) = @_;

	if ($args[0] =~ /=/) {
		if (@args == 1) {
			$self->{function} =
				   $self->state->functions->get('Mass action (irreversible)');
		} elsif (@args == 2) {
			$self->{function} =
				   $self->state->functions->get('Mass action (reversible)');
		} else {
			croak "invalid number of parameters\n";
		}

		return @args;
	}

	$self->{function} = $self->state->functions->get($args[0]);
	croak "unknown function name '$args[0]'\n" unless defined $self->function;
	shift @args;
	return @args;
}

sub _analyze_params {
	my ($self, @args) = @_;

	my $constants = $self->function->constants;
	croak   "invalid number of constants for function '" . $self->function->name
		  . "', needs " . @{$constants} . "\n" if @args != @{$constants};

	for (@args) {
		my @parts = split /\s*=\s*/, $_;
		croak "invalid parameter '$_'\n" unless @parts == 2;

		my $found = 0;
		for my $c (@{$constants}) {
			if ($parts[0] eq $c->name) {
				$found = 1;
				last;
			}
		}
		croak   "function '${\$self->function->name}' has no paramter "
			  . "named '$parts[0]'\n" unless defined $found;

		push @{$self->constants}, Param->new($self, $_);
	}
}


1;
