#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

package Report;
use v5.014;
use strict;
use warnings;
use Cwd 'abs_path';
use Carp;

use lib '.';
use parent 'Object';


sub new {
	shift->SUPER::new(state      => shift,
					  file_name  => undef,
					  species    => [],
					  with_rates => 0);
}


sub file_name {
	my $self = shift;
	$self->{file_name} = $self->state->model->name
		unless defined $self->{file_name};
	return $self->{file_name} . $self->state->iterator->desc;
}


sub set_with_rates {
	my $self = shift;
	$self->{with_rates} = shift != 0;
}


sub add {
	my ($self, $desc) = @_;
	my $species = $self->state->species->get($desc);
	if (defined $species) {
		push @{$self->species}, $species;
	} elsif (! defined $self->{file_name}) {
		$self->{file_name} = $desc;
	} else {
		croak "unknown species '$desc'\n";
	}
}


# Returns the index of a species in the list of reported species. If
# $rate is set skip species that are fixed or assigned.

sub index {
	my ($self, $species, $rate) = @_;
	my $i = 0;
	for (@{$self->species}) {
		return $i if $_ == $species;
		$i++ if    ! $rate
			    || (ref($_) ne 'Species_Assignment' && $_->type ne 'fixed');
	}

	return -1;
}


sub emit {
	my ($self, $indent) = @_;

	my $s .= "$indent  <ListOfReports>\n";
	$indent .= '  ';

	$s .=   "$indent<Report key=\"Report_0\" "
          . "name=\"Time and all Variable Values (Concentration Units)\" "
		  . "taskType=\"timeCourse\" separator=\", \" precision=\"6\">\n";
	$indent .= '  ';

	$s .= "$indent<Table printTitle=\"1\">\n";
	$indent .= '  ';

	$s .= sprintf "$indent<Object cn=\"CN=Root,Model=%s,Reference=Time\"/>\n",
	              $self->state->model->name;

	# If list of species to report is empty fill it with all species that
	# don't have a fixed concentration

	push @{$self->species}, grep({$_->type ne 'fixed'}
	                             @{$self->state->species->list})
	    if @{$self->species} == 0;

	$s .= sprintf   "$indent<Object cn=\"CN=Root,Model=%s,"
		          . "Vector=Compartments[compartment]," 
				  . "Vector=Metabolites[%s],"
				  . "Reference=Concentration\"/>\n",
				  $self->state->model->name, $_->name
		for @{$self->species};

	if ($self->with_rates) {
		for (@{$self->species}) {
			$s .= sprintf   "$indent<Object cn=\"CN=Root,Model=%s,"
			              . "Vector=Compartments[compartment]," 
						  . "Vector=Metabolites[%s],"
						  . "Reference=Rate\"/>\n",
						  $self->state->model->name, $_->name
				unless ref($_) eq 'Species_Assignment';
		}
	}

	$indent =~ s/  //;
	$s .= "$indent</Table>\n";

	$indent =~ s/  //;
	$s .= "$indent</Report>\n";

	$indent =~ s/  //;
	$s .= "$indent</ListOfReports>\n";
}


1;
