#!/usr/bin/perl

#  This file is part of the copd package
# 
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
# 
#  The program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
#  Copyright (C) 2019 Jens Thoms Törring <jt@toerring.de>
#

use v5.14;
use strict;
use warnings;


my $COPASI;
my $COPD;
my $GNUPLOT;


# Set paths to the COPASI program, the 'copd' directory and the
# gnuplot program from the environment and, if they're not defined
# to defaults.

BEGIN
{
	$COPASI  = $ENV{COPASI};
	$COPD    = $ENV{COPD};
	$GNUPLOT = $ENV{GNUPLOT};

	# Check if all are set or set defaults, first for Windows, then for Linux

	if ($^O eq 'MSWin32') {
		$COPASI = 'C:\Program Files\copasi.org\COPASI 4.24.197\bin\CopasiSE.exe'
			unless defined $COPASI;

		$COPD = 'C:\Users\pfruehwirt\Documents\AVL\Simulation\COPASI_Tool\copd'
			unless defined $COPD;

		$GNUPLOT = 'C:\Program Files\gnuplot\bin\gnuplot.exe'
			unless defined $GNUPLOT;
	} else {
		$COPASI = '/home/jens/Programming/Graz/COPASI/COPASI-4.23.184-Linux-64bit/bin/CopasiSE'
			unless defined $COPASI;

		$COPD = '/home/jens/Programming/Graz/COPASI/copd'
			unless defined $COPD;

		$GNUPLOT = '/usr/bin/gnuplot'
			unless defined $GNUPLOT;
	}
}


#######################################################
# Checks if the COPASI executable can be found (or at
# least something that looks like it)

sub check_copasi {
	my $copasi = shift;

	return if -r $copasi && -x $copasi;

	unless (   $copasi =~ /^\//
			|| ($^O eq 'MSWin32' && $copasi =~ /^[A-Z]:[\\\/]/)) {
		for (split /:/, $ENV{PATH}) {
			return if -r "$_/$copasi" && -x "$_/$copasi";
		}
	}

	say "Executable '$copasi' for program COPASI not found";
	exit -1;
}


#######################################################
# Checks if the COPASI executable can be found (or at
# least something that looks like it)

sub check_gnuplot {
	my $gnuplot = shift;

	return if -r $gnuplot && -x $gnuplot;

	unless (   $gnuplot =~ /^\//
			|| ($^O eq 'MSWin32' && $gnuplot =~ /^[A-Z]:[\\\/]/)) {
		for (split /:/, $ENV{PATH}) {
			return if -r "$_/$gnuplot" && -x "$_/$gnuplot";
		}
	}

	say "Execuble '$gnuplot' for program 'gnuplot not found";
	exit -1;
}


#######################################################
# Outputs file with data about fluoride emission rates
# in mol/(cm^2*h).

sub report_fer {
	my ($compartment, $filename, $type) = @_;

	# Get membrane thickness from compartment object (if this returns
	# 0 no thickness has been set)

	my $d = $compartment->thickness;
	if ($d == 0) {
		say "FER: no membrane thickness found in input file required for FER";
		return;
	}

	# Open COPASI report file which is supposed to contains the
	# HF production rate

	my $f;
	unless (open $f, '<', "$filename.crp") {
		say "FER: failed to open '$filename.crp' for reading";
		return;
	}

	# Find column containing the HF concentration or production rate

	my $line = <$f>;
	my $index = -1;
	my $rindex = -1;
	my $i = 0;
	my $search = $type eq 'full' ? 'HF.Rate' : '[HF]';
	for (split /,\s+/, $line) {
		$rindex = $i if $_ eq 'HF.Rate';
		$index = $i if $_ eq '[HF]';
		last if $index != -1 && ($type ne 'full' || $rindex != -1);
		$i++;
	}

	if ($index == -1 || ($type eq 'full' && $rindex == -1)) {
		say "FER: no HF concentration (or rate) found in file '$filename.crp'";
		close $f;
		return;
	}

	# Create file for the FER data, same name as COPASI file but with
	# '.fer' extension

	my $o;
	unless (open $o, '>', "$filename.fer") {
		say "FER: failed to open '$filename.fer' for writing";
		close $f;
		return;
	}

	if ($type eq 'full') {
		# Outptut the fluoride emission rate FER, which is calculated as
		# FER [mol/(cm^2*h)] = 3.6 * HF.Rate [mol/(L*s)] * d(Membran) [cm]

		print $o "Time, FER, avg(FER)\n";

		while (<$f>) {
			my @list = split /,\s+/, $_;
			print $o $list[0] . ", ". 360 * $list[$rindex] * $d;
			if ($list[0] != 0) {
				print $o ", " . (360 * $list[$index] * $d / $list[0]) . "\n";
			} else {
				print $o "\n";
			}
		}
	} else {
		print $o "Time, avg(FER)\n";

		while (<$f>) {
			my @list = split /,\s+/, $_;
			print $o   $list[0] . ", "
				     . (360 * $list[$index] * $d / $list[0]) . "\n"
				if $list[0] != 0;
		}
	}

	close $o;
	close $f;
}


#######################################################

sub parse_bad {
	die "Unexpected content at start of input file\n";
}


#######################################################

sub parse_model {
	shift->model->set_name(shift);
}


#######################################################

sub parse_compartment {
	my ($obj, $line) = @_;
	if ($line =~ /^\s*v(olume)?\s*[:=]\s*(.*)$/i) {
		$obj->compartment->set_volume($2);
	} elsif ($line =~ /^\s*t(hickness)?\s*[:=]\s*(.*)$/i) {
		$obj->compartment->set_thickness($2);
	} else {
		$obj->compartment->set_volume($line);
	}
}


#######################################################

sub parse_constants {
	shift->add_constant(shift);
}


#######################################################

sub parse_functions {
	shift->add_function(shift);
}


#######################################################

sub parse_species {
	shift->add_species(shift);
}


#######################################################

sub parse_reactions {
	shift->add_reaction(shift);
}


#######################################################

sub parse_task {
	shift->task->set_duration(shift);
}


#######################################################

sub parse_report {
	shift->report->add(shift);
}


sub parse_plot  {
	shift->plots->add_plot(shift);
}


#######################################################

sub usage {
	say "Missing input file name" if shift;
	say "Usage:  copd [-p] [-r] [-f] FILE";
	say "Options";
	say "  -p emit file with plot of data";
	say "  -r write rates to outout file";
	say "  -f calculate FER (full data set requires '-r')";
	exit 0;
}


#######################################################


use lib $COPD;

# Check that the COPASI executable is where we assume

check_copasi($COPASI);

use State;
my $state = State->new;

my %parsers = (Model       => \&parse_model,
			   Compartment => \&parse_compartment,
			   Constants   => \&parse_constants,
			   Functions   => \&parse_functions,
			   Species     => \&parse_species,
			   Reactions   => \&parse_reactions,
			   Task        => \&parse_task,
			   Report      => \&parse_report,
			   Plots       => \&parse_plot);

usage(1) unless @ARGV;;

my $plot = 0;
my $with_rates = 0;
my $with_fer;
my $filename;

my $arg;
while ($arg = shift @ARGV) {
	if ($arg eq '-h' || $arg eq '--help') {
		usage(0);
	} elsif ($arg eq '-p') {
		$plot = 1;
	} elsif ($arg eq '-r') {
		$with_rates = 1;
		$state->report->set_with_rates(1);
	} elsif ($arg eq '-f') {
		$with_fer = 'avg';
	} elsif (@ARGV == 0) { 
		$filename = $arg;
	} else {
		usage(1);
	}
}

usage(1) unless defined $filename;


check_gnuplot($GNUPLOT) if $plot;


if ($with_fer && $with_rates) {
	$with_fer = 'full';
}

# Try to open the input file

open my $file, '<', $filename
	or die "Can't open file '$filename' for reading\n";

# Read file in line by line and parse them, switching parsers at lines
# that indicate the start of a new section

my $parse = \&parse_bad;

while (defined (my $line = <$file>)) {
	# Remove comments and leading and trailin white space, skip empty lines

	$line =~ s/^\s*(.*?)(\s*#.*?)?\s*$/$1/;
	next if $line eq '';

	# Check if the line is a section header and switch parser accordingly,
	# then get the next line

	if ($line =~ /^([^:]+)\s*:$/ && exists $parsers{$1}) {
		$parse = $parsers{$1};
		next;
	}

	# Try to parse the line, use eval to be able to add information
	# about the current line on errors

	eval {$parse->($state, $line)};
	if ($@) {
		my $mess = $@;
		$mess =~ s/\n+.*$//;
		chomp $mess;
		say "$mess at input line $.: \"$line\"";
		exit -1;
	}
}

close $file;

$_->check for @{$state->functions->list};
$state->plots->check($with_rates);

# Run the calculation(s) - we end when the iterator has reached the end
# and its next method returns undef

my $ret = 0;
do {
	# Create the 'cps' file

	$filename = $state->report->file_name;
	open $file, '>', $filename . '.cps'
		or die "Can't open file '$filename.cps' for writing\n";
	print $file $state->emit;
	close $file;

	# Run COPASI on that file (make sure both the executable and the file
	# name are supplied in double quotes to avoid issues with spaces in
	# the names!)

	say "Running '$filename.cps'";
	`"$COPASI" "$filename.cps"`;
	$ret = $?;
	if ($ret) {
		say "COPASI failed on '$filename.cps'";
	} else {
		report_fer($state->compartment, $filename, $with_fer) if $with_fer;
		$state->plots->gnuplot_plot($filename, $GNUPLOT) if $? == 0 && $plot;
	}
} while (defined $state->iterator->next);

exit $ret;
